; Check for tile move, for The Wall.
; copyright (C) W. Lenerz 2000-2016, see the accompanying licence_txt file.


	section chk_moves
	include dev1_asm_wall_includes
	include dev1_asm_incl_keys_err

; this checks whether some tiles should move after user clicked somewhere

; on entry,			on exit
; D0 = row nbr			????
; D1 = column nbr		????
; D5 = ????			????
; A2 = ????			pointer to (new) data
; condition codes set on exit: if Z, nothing moved
; if NE, something moved

	xdef	chk_move
	xdef	chk_poss
	xref	sound

chkregs REG	A1/A2/D4/D7

chk_move	MOVEM.L chkregs,-(A7)
	MOVEQ	#0,D5			; D5 = flag, if <>0 at end, there were connecting ones
	MOVE.L	current(A4),A2		; point to current state of board
	MOVEQ	#0,D2
	MOVE.W	D0,D2			; d2 = row number
	MULU	#nbr_cols,D2
	ADD.W	D1,D2			; D2 now points to element in (A2) table where we clicked
	MOVE.B	(A2,D2.W),D7		; D7 = our type of tile
	BNE.S	not_nul 		; ok, check it->...
	MOVEQ	#0,D0			;
	BRA.S	do_sound		; there is no tile at that space; so
					; complain (sound) & return->...

; this creates a new tiles record in the table, which will be pointed to by
; current(A4)
not_nul ADD.L	#tottiles,A2		; next tiles record
	MOVE.L	moves(A4),A1
	ADD.L	#totmem-tottiles,A1 ; end of available memory
	MOVEQ	#err.orng,D4		; preset error
	CMP.L	A1,A2			; new tiles record mustn't exceed this!
	BGT.S	move_out		; oops! this should NEVER happen!
	MOVE.L	current(A4),A1		; point to current state of board
	MOVE.L	A2,-(A7)		; keep next state of board
	MOVE.W	#tottiles/2-1,D4	; number of tiles /2 for word sized tranfser
mvloop	MOVE.W	(A1)+,(A2)+		; transfer board state (as words)
domvlp	DBF	D4,mvloop
	MOVEM.L (A7)+,A2		; a2 points to new current tiles record
	BSR.S	do_check		; check moves now
	MOVE.W	D5,D0			; were any tiles taken away?
	BEQ.S	do_sound		; no->...
	MOVE.L	A2,current(A4)		; yes, so this is current state of board now
	BSR	tilefall		; check whether tiles should move down
do_sound
	JSR	sound			; make a sound (use the "error" return in D5->D0!)
move_out
	MOVEM.L (A7)+,chkregs		; get regs back
	TST.W	D5			; did any tiles move?
	RTS


; on entry,			on exit
; D1 = column nbr			???
; D0 = row nbr			????
; D2 = where we are with respect to A2	????
; D5 = 0				if 0 : no tiles could be taken away,
;				if 1: tiles were taken away
; A2 = ptr to START OF data		unchanged
		       

; do_check: this is a recursive routine
; it checks, for the tile hit, whether there are matching neighbouring tiles
; if yes, it checks those tiles, too
doregs	REG	D1-D2
do_check  MOVEM.W doregs,-(A7)
chk_up	MOVE.W	D2,D3			; keep pointing to myself
	TST.W	D0			; upper row reached?
	BEQ.S	chk_down		; yes, so we can't go up->
	MOVE.B	-nbr_cols(A2,D2.W),D6 ; get the tile from the row above
	CMP.B	D6,D7			; is it the same?
	BNE.S	chk_down		; no->...
	SUBQ.W	#1,D0			; yes, new row
	MOVE.B	#-1,(A2,D2.W)		; show that there is a connecting tile
	SUB.W	#nbr_cols,D2		; new pointer within data
	MOVE.B	#-1,(A2,D2.W)		; this is also a connecting tile,(the two tiles are the same, so they are "connected")
	MOVEQ	#1,D5			; show that tiles must be taken away
	BSR.S	do_check		; recurse !
	ADDQ.W	#1,D0			; point to my row again
	ADD.W	#nbr_cols,D2	;	 point to my tile again
; check the tile in the next row
chk_down
	CMP.W	#nbr_rows-1,D0		; are we in last row?
	BEQ.S	chk_left		; yes, so do nothing here ->...
	MOVE.B	nbr_cols(A2,D2.W),D6	; get the tile from next row
	CMP.B	D6,D7			; is it the same?
	BNE.S	chk_left		; no->...
	ADDQ.W	#1,D0			; yes, new row
	MOVE.B	#-1,(A2,D2.W)		; show that there is a connecting one
	ADD.W	#nbr_cols,D2		; new pointer within data
	MOVE.B	#-1,(A2,D2.W)		; this is also a connecting one
	MOVEQ	#1,D5
	BSR.S	do_check		; recurse!
	SUBQ.W	#1,D0
	SUB.W	#nbr_cols,D2
; check the tile to the left
chk_left	TST.W	D1		; are we in first col?
	BEQ.S	chk_right		; yes, so do nothing here ->...
	MOVE.B	-1(A2,D2.W),D6		; get the one from column to the left
	CMP.B	D6,D7			; is it the same?
	BNE.S	chk_right		; no->...
	SUBQ.W	#1,D1			; new column
	MOVE.B	#-1,(A2,D2.W)		; show that there is a connecting one
	SUBQ.W	#1,D2			; new pointer within data
	MOVE.B	#-1,(A2,D2.W)		; this is also a connecting tile
	MOVEQ	#1,D5
	BSR	do_check		; recurse !
	ADDQ.W	#1,D1
	ADD.W	#1,D2
; check the tile to the right
chk_right CMP.W #nbr_cols-1,D1		; are we in last col?
	BEQ.S	chk_out 		; yes, so do nothing here ->...
	MOVE.B	1(A2,D2.W),D6		; get the one from col to the right
	CMP.B	D6,D7			; is it the same?
	BNE.S	chk_out 		; no->...
	ADDQ.W	#1,D1			; new col
	MOVE.B	#-1,(A2,D2.W)		; show that there is a connecting one
	ADDQ.W	#1,D2			; new pointer within data
	MOVE.B	#-1,(A2,D2.W)		; this is also a connecting one
	MOVEQ	#1,D5
	BSR	do_check		; recurse!
chk_out MOVEM.W (A7)+,doregs
	TST.W	D5			; found any?
	RTS
 


; this checks whether the tiles in a column should be moved down since some
; disappeared underneath
; (for each row, scan each col, until either top row, or empty tile, reached)
; this also eliminates totally empty columns
tfregs	REG	A1-A3/D0/D3-D5
tilefall	MOVEM.L tfregs,-(A7)
	MOVE.L	current(A4),A2		; A2 points to current tiles record
	MOVE.L	A2,D3
	ADD.L	#nbr_cols,D3		; point to start of row 2
	ADD.L	#tottiles-nbr_cols,A2
	MOVEQ	#0,D0
	MOVE.W	lastcol(A4),D0
	ADD.L	D0,A2
	MOVE.L	A2,A1
; inner loop, within one column
tfloop1 MOVE.B	(A2),D2 		; get this tile
	BEQ.S	end_col 		; it's empty, so end of scan for this row
	BGT.S	do_tflp1		; it's a normal tile, do nothing
	MOVE.B	#0,(A2) 		; just to be sure
	MOVE.L	A2,A3
	BRA.S	mvd2
movedown
	SUB.L	#nbr_cols,A3		; point to row above
mvd2	CMP.L	D3,A3			; is this the upper row?
	BGE.S	ddown			; no->...
	MOVE.B	#0,(A3) 		; yes, so end this run
	BRA.S	tfloop1
ddown	MOVE.B	-nbr_cols(A3),(A3)
	BRA.S	movedown
do_tflp1
	SUB.L	#nbr_cols,A2		; point to next row
	CMP.L	D3,A2			; are we in upper row (special case?)
	BGE.S	tfloop1 		; no, so go around again->...
upperrow
	TST.B	(A2)			; upper row, if matched, just set to 0
	BGE.S	end_col
	MOVE.B	#0,(A2)
end_col CMP.L	A1,A2			; is the entire column empty?
	BNE.S	no_join 		; no->...
	MOVE.L	A2,D4			;
	SUB.L	current(A4),D4
	SUB.L	#tottiles-nbr_cols,D4	; D4 = column number
	MOVE.W	lastcol(A4),D2
	SUB.W	D4,D2			; delete last row?; res in D2
	BEQ.S	one_less		; yes, nothing much to do
	SUBQ.W	#1,D2			; for DBF
	MOVE.W	#nbr_rows-1,D5		; one inner loop for each row
collp	MOVE.W	D2,D1			; how often per row
	MOVE.L	A2,-(A7)
rowcpy	MOVE.B	1(A2),(A2)+		; shuffle down
	MOVE.B	#0,(A2) 		;
	DBF	D1,rowcpy		; as often as necessary
	MOVE.L	(A7)+,A2
	SUB.L	#nbr_cols,A2		; up to next row
	DBF	D5,collp
one_less
	SUBQ.W	#1,lastcol(A4)
no_join SUBQ.L	#1,A1			; do it for all rows
	MOVE.L	A1,A2
do_tflp2
	DBF	D0,tfloop1
	MOVEQ	#nbr_cols-1,D0
	MOVE.L	current(A4),A2
tfloop3 MOVE.B	(A2)+,D1
	BGE.S	do_tflp3
	MOVE.B	#0,-1(A2)
do_tflp3
	DBF	D0,tfloop3
	MOVEM.L (A7)+,tfregs
	RTS


; this checks whether moves are still possible
; see whether a matching tiles is found to the right, or down
; don't need to check uo & left)
; on exit : D0 = 0 if moves are possbile
;	  D0 = 1 if no more move are possbile (i.e. end of game)
;	  D0 = -1 if game won
; cond codes set on exit

possregs	REG	D1-D7/A2
chk_poss
	MOVEM.L possregs,-(A7)
	MOVEQ	#0,D0
	MOVEQ	#0,D1
	MOVEQ	#0,D2
	MOVEQ	#0,D3
	MOVE.L	current(A4),A2
	MOVEQ	#nbr_rows-2,D5
in_loop MOVEQ	#nbr_cols-2,D4
poss_rite
	MOVE.B	(A2),D7 		; tile currently pointed to
	BEQ.S	none			; there is no tile
	MOVEQ	#-1,D3			; flag that there are still tiles
	MOVE.B	1(A2),D6		; get the one from next col
	CMP.B	D6,D7			; is it the same?
	BEQ.S	found			; yes!
poss_down
	MOVE.B	nbr_cols(A2),D6 	; get the one from next row
	CMP.B	D6,D7			; is it the same?
	BEQ.S	found			; yes!
none	ADDQ.L	#1,A2			; point to next tile in column
	DBF	D4,poss_rite		; do it for n-1 columns
; here we point to the last column, no need to check for right
	MOVE.B	(A2),D7
	BEQ.S	none2
	MOVEQ	#-1,D3
	MOVE.B	nbr_cols(A2),D6 	; get the one from next row
	CMP.B	D6,D7			; is it the same?
	BEQ.S	found			; yes!
none2	ADDQ.L	#1,A2			; point to next
	DBF	D5,in_loop
; here we are in last row, no need to check for down
	MOVEQ	#nbr_cols-2,D4
pright	MOVE.B	(A2),D7 		; tile currently pointed to
	BEQ.S	none3			; there is not tile
	MOVEQ	#-1,D3
	MOVE.B	1(A2),D6		; get the one from next col
	CMP.B	D6,D7			; is it the same?
	BEQ.S	found			; yes!
none3	ADDQ.L	#1,A2
	DBF	D4,pright
	MOVEQ	#1,D0			; no more are possbible
	TST.W	D3			; but are there any tiles at all?
	BNE.S	possout 		; yes, so game is lost->...
	MOVEQ	#-1,D0			; no, so game is won!
possout MOVEM.L (A7)+,possregs
	TST.L	D0			; found any?
	RTS
found	MOVEQ	#0,D0			; there are still some moves
	BRA.S	possout

	end
